from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, List, Optional

from pydantic import BaseModel, Field, root_validator
from wildered.logger import logger
from wildered.ast import ASTSourceCode
from wildered.directive import Identifier

from .directives import HintDirective


class Dependency(BaseModel, ABC):
    @abstractmethod
    def resolve(self) -> str:
        raise NotImplementedError


class CodeDependency(Dependency):
    """
    Represents a code dependency that can be resolved and used within the project.
    """

    filepath: Path  # Where to find the code
    entity_name: str = ""
    source_code: ASTSourceCode

    @root_validator(pre=True)
    def initialize_source_code(cls, v):
        v.setdefault(
            "source_code", ASTSourceCode.from_file(filename=str(v["filepath"]))
        )
        return v

    def __eq__(self, __value: CodeDependency) -> bool:
        if not isinstance(__value, CodeDependency):
            raise ValueError("CodeDependency can only compare with its own type")
        return (__value.filepath == self.filepath) and (
            self.entity_name == __value.entity_name
        )

    def resolve(self) -> str:
        if self.entity_name:
            return self.source_code.get_entity(
                entity_name=self.entity_name,
                drop_directive=True,
                directive_prefix="wildered",
                return_global_import=True,
            )

        else:
            return self.source_code.unparse()


class ReferenceDependency(Dependency):
    index_name: str  # What is the name of the vectorstore?
    query: str  # What is the keyword?
    filter_criteria: Any  # Any other filtering criteria?


def infer_hint_list(hint_list: List[HintDirective], source: ASTSourceCode):
    # All should point to the same ASTSourceCode

    dependency_list = []
    for hint in hint_list:
        cur_list = infer_hint(
            hint_directive=hint,
            source=source,
            dependency_lookup=source.get_entity_map(),
        )
        dependency_list.extend(cur_list)
    return dependency_list


def infer_hint(
    hint_directive: HintDirective, source: ASTSourceCode, dependency_lookup: dict
) -> List[Dependency]:
    logger.debug(f"Receiving {dependency_lookup=} and {hint_directive=}")
    code_dependency = []

    for entity in hint_directive.entity_list:
        match entity:
            case Identifier():
                try:
                    if module_path := dependency_lookup.get(entity.name, None):
                        code_dependency.append(
                            CodeDependency(
                                filepath=module_path,
                                entity_name=entity.name,
                            )
                        )

                    else:
                        code_dependency.append(
                            CodeDependency(
                                filepath=source.filename,
                                entity_name=entity.name,
                            )
                        )
                        
                except FileNotFoundError as e:
                    logger.debug(f"{entity.name=} not found")

            case str():
                component = entity.split(":")
                filepath, entity_name = (component[0], component[1])
                logger.debug(f"Specifying dependencies in raw string: {filepath=}, {entity_name=}")
                filepath = apply_relative_path(
                    relative_path=filepath,
                    absolute_base_path=source.filename.parent.resolve(),
                )
                try:
                    code_dependency.append(
                        CodeDependency(filepath=filepath, entity_name=entity_name)
                    )
                except FileNotFoundError as e:
                    logger.debug(f"Cannot find {filepath=}")

            case other:
                raise TypeError(f"Unknown type {type(other)}")

    return code_dependency


def apply_relative_path(relative_path: Path, absolute_base_path: Path) -> Path:
    """
    Applies a relative path to a base path, similar to the behavior of the 'cd' command in Linux.

    Args:
        base_path (Path): The base path to which the relative path will be applied.
        relative_path (Path): The relative path to be applied.

    Returns:
        Path: The resulting path after applying the relative path to the base path.

    """
    result_path = (
        absolute_base_path / relative_path
    )  # Apply the relative path to the base path
    return result_path.resolve()


def union_dependencies(dependencies: List[Dependency]) -> str:
    cur_dependencies = []
    for i in dependencies:
        for j in cur_dependencies:
            if i == j:
                break
        else:
            # TIPS: Will get here if there is no break
            cur_dependencies.append(i)

    return cur_dependencies



from typing import List

from wildered.ast import ASTSourceCode

from ..autocomplete import task_executor
from ..directives import butterfly_parser
from ..tasks import TaskGroup, task_grouper


def scan(
    filename: str,
    clipboard: bool = False,
    remove_directive: bool = False,
    auto_integrate: bool = False,
) -> None:
    source_code = ASTSourceCode.from_file(filename)
    task_groups = _get_task_groups(source=source_code)
    if task_groups:
        task_executor(
            task_groups,
            clipboard=clipboard,
            auto_integrate=auto_integrate
        )
        
        if remove_directive:
            # Since directive is already removed during parsing
            source_code.save()

    else:
        print("No butterfly directive detected.")


def _get_task_groups(source: ASTSourceCode) -> List[TaskGroup]:
    entity_list = butterfly_parser.parse(source=source, drop_directive=True)
    entity_groups = task_grouper.group(entity_list=entity_list)
    task_groups = [TaskGroup.from_entity_group(i) for i in entity_groups]
    return task_groups



import sys
from pathlib import Path

from loguru import logger

CLEAR_LOG_ON_EVERY_RUN = True

if CLEAR_LOG_ON_EVERY_RUN:
    if (x := Path("debug.log")).is_file():
        x.unlink()

logger.remove(0)
logger.level("INFO", color="<white>")
logger.level("WARNING", color="<light-red>")

logger.add(
    sink="debug.log", level="DEBUG", rotation="00:00", retention="1 day", colorize=False
)

logger.add(
    sink=sys.stderr,
    level="INFO",
)


def autocomplete(
    *,
    requirement=None,
    rewrite_docs: bool = False,
    annotation: bool = True,
    count: int = 1,
    group: str = None,
    pseudo_code: bool = False,
    modify: bool = False,
):
    """
    If placed on the class this means autocomplete the whole class (allow changes of the everything in the class)
    You can specify how would you want to complete the class through the requirements.

    If placed on the class method, it allow changes of the class implementation. (The difference with class is that
    only changes to that specific class method will be applied)

    If placed on function, will only make change to that specific function.

    Arguments:
    - requirement: How to implement the entity
    - group: Specify which to be implemented together, often used when two todo functions are interrelated.
    """
    return


def hint(*, entity: list, infer: bool = False):
    """
    Declare that this task will depend on some specific functions and class
    if module is None, then by default it is the current module
    - infer will get additional dependencies on top of the already specified hint
    """


__all__ = ["hint", "autocomplete"]



from typing import List, Optional

import libcst as cst


class CSTDropImplementation(cst.CSTTransformer):
    def __init__(self, exception: List[str]) -> None:
        self.exception = exception

    def visit_FunctionDef(self, node: cst.FunctionDef):
        func_name = node.name.value
        if func_name in self.exception:
            return node

        else:
            docstring = node.get_docstring()
            if docstring:
                # node.body = [cst.Expr(value=cst.Constant(value=docstring))]
                # TODO
                pass

            else:
                node.body = []

            return node
        
        
class CSTDropDirective(cst.CSTTransformer):
    def __init__(self, prefix: str):
        self.prefix = prefix

    def visit_FunctionDef(self, node: cst.FunctionDef) -> cst.FunctionDef:
        node.decorators = filter(self.filter_directive, node.decorators)
        return node

    def visit_ClassDef(self, node: cst.ClassDef) -> cst.ClassDef:
        node.decorators = filter(self.filter_directive, node.decorators)
        return node

    def visit_Expr(self, node: cst.Expr) -> Optional[cst.Expr]:
        try:
            prefix, node_name = get_call_name(node.value)
            if (prefix == self.prefix) and (node_name == "run"):
                return None

        except AttributeError:
            return node

    def visit_Import(self, node: cst.Import) -> Optional[cst.Import]:
        if node.names[0].name.value.startswith(self.prefix):
            return None

        else:
            return node

    def visit_ImportFrom(self, node: cst.ImportFrom) -> Optional[cst.ImportFrom]:
        if node.module.value.startswith(self.prefix):
            return None

        else:
            return node

    def filter_directive(self, decorator: cst.Decorator) -> bool:
        decorator_obj = decorator.decorator
        if isinstance(decorator_obj, cst.Name):
            return True

        prefix, node_name = get_call_name(decorator_obj)

        if prefix == self.prefix:
            return False

        return True    

def get_call_name(node: cst.Call) -> str:
    func_obj = node.func
    if hasattr(func_obj, "attr"):
        return (func_obj.value.value, func_obj.attr.value)

    else:
        return ("", func_obj.value)


from __future__ import annotations

import copy
from pathlib import Path
from typing import Any, List, Optional

import libcst as cst

from wildered.cst.utils import CSTDropDirective, CSTDropImplementation
from wildered.models import BaseSourceCode
from wildered.utils import read_file, write_file


class CSTSourceCode(BaseSourceCode):
    node: Any # Bad things happen when I change to cst.Module
    filename: Path        
    
    def __str__(self) -> str:
        return self.node.code
    
    def unparse(
        self,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        exception: Optional[List[str]] = None,
    ) -> str:
        return self._unparse(
            node=self.node,
            directive_prefix=directive_prefix,
            drop_directive=drop_directive,
            drop_implementation=drop_implementation,
            exception=exception,
            return_global_import=False,
        )

    def _unparse(
        self,
        node: cst.Module,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        exception: Optional[List[str]] = None,
        return_global_import: bool = False,
    ) -> str:
        # We need to create a copy to avoid mutation
        node = copy.deepcopy(node)
        if drop_directive:
            if directive_prefix == "":
                raise ValueError(
                    "Must specify a directive_prefix when drop_directive set to True"
                )
            transformer = CSTDropDirective(prefix=directive_prefix)
            node = transformer.visit(node)

        if drop_implementation:
            transformer = CSTDropImplementation(exception=exception if exception else [])
            node = transformer.visit(node)

        if return_global_import:
            return (
                self.get_import_statement(
                    drop_directive=drop_directive, directive_prefix=directive_prefix
                )
                + "\n\n"
                + node.code
            )

        else:
            return node.code
    
    # def get_import_statement(
    #     self,
    #     drop_directive: bool = False,
    #     directive_prefix: str = "",
    #     relative_only: bool = False,
    # ) -> str:
    #     """
    #     Return the leading import statements (statements that are placed at the top of the file)
    #     in string format
    #     """
    #     import_list = extract_import_list(
    #         self.node,
    #         drop_directive=drop_directive,
    #         directive_prefix=directive_prefix,
    #         relative_only=relative_only,
    #     )
    #     import_strings = [cst.unparse(node) for node in import_list]

    #     # Combine the import strings into a single string
    #     return "\n".join(import_strings)
    
    @classmethod
    def from_file(cls, filename: str) -> CSTSourceCode:
        content = read_file(filename)
        node = cst.parse_module(content)
        return cls(node=node, filename=filename)
    
    def save(self, filename: Optional[str] = None) -> None:
        filename = filename if filename else self.filename
        write_file(filename, self.node.code)


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional
from uuid import uuid4

from pydantic import BaseModel, PrivateAttr, validator
from typing_extensions import assert_never

from wildered.ast.directive_parser import (
    ASTClassEntity,
    ASTFunctionEntity,
    ASTModuleEntity,
    BaseEntity,
)
from wildered.group import EntityGroup, EntityGrouper
from wildered.models import BaseSourceCode
from wildered.logger import logger

from .dependency import Dependency, infer_hint_list, union_dependencies
from .directives import AutocompleteDirective
from .prompts import CODER_PROMPT
from .utils import wander_render


def get_group_name(entity: BaseEntity) -> str:
    if group_name := entity.directives["autocomplete"][0].group:
        return group_name

    else:
        return str(uuid4())


task_grouper = EntityGrouper(group_func=get_group_name)


class Task(BaseModel):
    task_type: Literal["class", "function", "file"]
    entity_name: Optional[str] = None
    dependencies: List[Dependency] = []
    source: BaseSourceCode
    node: BaseEntity

    @classmethod
    def from_entity(cls, node: BaseEntity) -> Task:
        task_type = ""
        match node:
            case ASTClassEntity():
                task_type = "class"

            case ASTFunctionEntity():
                task_type = "function"

            case ASTModuleEntity():
                task_type = "file"

            case other:
                assert_never(other)

        if node.directives.get("hint", False):
            dependency_list = infer_hint_list(
                hint_list=node.directives["hint"], source=node.source
            )

        else:
            dependency_list = []

        return cls(
            task_type=task_type,
            entity_name=node.name,
            dependencies=dependency_list,
            source=node.source,
            node=node,
        )

    def return_task_context(self) -> str:
        if self.nested:
            top_level_node = self.node.ancestor[-1]
            return top_level_node.unparse()

        else:
            return self.node.unparse()
        
    def integrate(
        self,
        new_code: str,
        **kwargs
    ) -> None:
        self.node.update(new_code=new_code, **kwargs)
        self.node.save()
        

    @property
    def requirement(self):
        autocomplete_config: AutocompleteDirective = self.node.directives[
            "autocomplete"
        ][0]
        requirements = ""
        if autocomplete_config.modify:
            match self.task_type:
                case "file":
                    brief = (
                        f"Modify the program called {self.source.filename.name}. "
                    )

                case "function":
                    brief = f"Modify existing implementation in {self.entity_name} function. "

                case "class":
                    brief = (
                        f"Modify existing implementation in {self.entity_name} class. "
                    )

                case other:
                    assert_never(other)

            if not autocomplete_config.requirement:
                brief += "Follow instruction from comments starting with 'm:' on how to modify the implementation. "

            requirements = brief + autocomplete_config.requirement

        else:
            match self.task_type:
                case "file":
                    brief = f"Rewrite the {self.source.filename.name}.py script. "

                case "function":
                    brief = f"Implement {self.entity_name} function. "

                case "class":
                    brief = f"Implement {self.entity_name} class. "

                case other:
                    assert_never(other)

            if not autocomplete_config.requirement:
                brief += f"Follow instruction from docstrings and comments on how to implement the {self.task_type}. "

            requirements = brief + autocomplete_config.requirement

        if autocomplete_config.pseudo_code:
            requirements = (
                requirements
                + "You may implement the solution based on the pseudo-code specified in the docstrings or comments. "
            )

        if autocomplete_config.rewrite_docs:
            requirements = (
                requirements
                + "You are required to rewrite the docstrings to follow the Python pandas library docstrings style. "
            )

        if autocomplete_config.annotations:
            requirements = (
                requirements + "Ensure that your answer has type annotations. "
            )

        requirements = wander_render(requirements)

        return requirements

    @property
    def nested(self) -> bool:
        return self.node.wrapping_node is not None

    @property
    def group_name(self) -> Optional[str]:
        return self.node.directives["autocomplete"][0].group


def get_additional_context(dependencies: List[Dependency]) -> str:
    if dependencies != []:
        result = "Below are snippets of code in the current project that you may find useful:\n"
        for dependency in dependencies:
            result += "```python\n" + dependency.resolve() + "\n```" + "\n"
        return result

    else:
        return ""


class TaskGroup(BaseModel):
    group_name: Optional[str] = None
    task_list: List[Task]
    _aggregate_dependencies: List[Dependency] = PrivateAttr(default_factory=list)

    @property
    def aggregate_dependencies(self) -> List[Dependency]:
        if self._aggregate_dependencies == []:
            dependency_list = []
            for i in self.task_list:
                dependency_list.extend(i.dependencies)
            self._aggregate_dependencies = union_dependencies(
                dependencies=dependency_list
            )
        return self._aggregate_dependencies

    @classmethod
    def from_entity_group(cls, entity_group: EntityGroup) -> TaskGroup:
        task_list = [Task.from_entity(entity) for entity in entity_group.entity_list]
        return TaskGroup(group_name=entity_group.group_name, task_list=task_list)

    def format_prompt(self, template=CODER_PROMPT) -> str:
        total_requirement = ""
        total_code_context = ""
        total_add_context = ""
        for i, task in enumerate(self.task_list):
            total_code_context += task.return_task_context()
            additional_context = get_additional_context(task.dependencies)
            total_add_context += additional_context
            total_requirement += f"## Task {i + 1}\n" + task.requirement + "\n"

        input_dict = {
            "requirement": total_requirement,
            "code_context": total_code_context,
            "additional_context": total_add_context,
        }

        formatted_prompt = template.format(**input_dict)
        return formatted_prompt
    
    def integrate(self, response: str) -> None:
        for i in self.task_list:
            i.integrate(new_code=response)
        
        # We only have to update the import statement once
        self.task_list[0].source.update_import_statement(new_code=response)



from __future__ import annotations

import ast
import copy
import pickle
from pathlib import Path
from typing import Annotated, Dict, List, Optional, TypeAlias

import ast_comments
from pydantic import PrivateAttr

from wildered.ast.utils import (
    DropDirective,
    DropImplementation,
    ReplaceNode,
    diff_import_list,
    extract_import_list,
    locate_class,
    locate_entity,
    locate_function,
    locate_method,
)
from wildered.models import BaseSourceCode
from wildered.utils import read_file, resolve_module_filepath, write_file

EntityMap: TypeAlias = Dict[
    Annotated[str, "The entity name"],
    Annotated[str, "The path of the file containing the entity"],
]


class ASTSourceCode(BaseSourceCode):
    """
    A class that parses a given Python script and facilitates mmanipulation of components
    such as function and docstrings in the script.
    The main way of initializing this is by using `from_file` method
    """

    node: ast.AST
    filename: Optional[Path] = None
    _entity_map: Optional[EntityMap] = PrivateAttr(default=None)

    def get_import_statement(
        self,
        drop_directive: bool = False,
        directive_prefix: str = "",
        relative_only: bool = False,
    ) -> str:
        """
        Return the leading import statements (statements that are placed at the top of the file)
        in string format
        """
        import_list = extract_import_list(
            self.node,
            drop_directive=drop_directive,
            directive_prefix=directive_prefix,
            relative_only=relative_only,
        )
        import_strings = [ast_comments.unparse(node) for node in import_list]

        # Combine the import strings into a single string
        return "\n".join(import_strings)

    def __str__(self) -> str:
        return ast_comments.unparse(self.node)

    def update(self, class_to_replace, method_to_replace, function_to_replace):
        replacer = ReplaceNode(
            class_to_replace=class_to_replace,
            function_to_replace=function_to_replace,
            method_to_replace=method_to_replace,
        )
        self.node = replacer.visit(self.node)

    def update_function(self, new_code: str | ast.FunctionDef, func_name: str) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)
            new_code = locate_function(ast_obj=new_code, func_name=func_name)

        replace_transformer = ReplaceNode(function_to_replace={func_name: new_code})
        replace_transformer.visit(self.node)
        ast.fix_missing_locations(self.node)

    def update_method(self, new_code: str, func_name: str, class_name: str) -> None:
        new_code = ast_comments.parse(new_code)
        try:
            new_method = locate_method(
                ast_obj=new_code, func_name=func_name, class_name=class_name
            )
            replace_transformer = ReplaceNode(
                method_to_replace={class_name: {func_name: new_method}}
            )
        except ValueError:
            new_method = locate_function(ast_obj=new_code, func_name=func_name)
            replace_transformer = ReplaceNode(
                method_to_replace={class_name: {func_name: new_method}}
            )
        self.node = replace_transformer.visit(self.node)
        ast.fix_missing_locations(self.node)

    def update_class(self, new_code: str, class_name: str) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)
            new_code = locate_class(ast_obj=new_code, class_name=class_name)

        replace_transformer = ReplaceNode(class_to_replace={class_name: new_code})
        replace_transformer.visit(self.node)
        ast.fix_missing_locations(self.node)

    def update_module(self, new_code: str | ast.Module) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)

        self.node = new_code

    def update_import_statement(self, new_code: str) -> None:
        """
        import_list should be a list of ast node
        replace the import lines in self.code to the one in import_list
        """
        code = ast_comments.parse(new_code)
        existing_import_list = extract_import_list(self.node)
        new_import_list = extract_import_list(code)
        new_import = diff_import_list(
            existing_import_list, new_import_list
        )
        for i in new_import:
            self.node.body.insert(0, i)

    def get_function(
        self,
        func_name: str,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        return_global_import: bool = False,
    ) -> str:
        function_node = locate_function(ast_obj=self.node, func_name=func_name)
        return self._unparse(
            node=function_node,
            drop_directive=drop_directive,
            directive_prefix=directive_prefix,
            drop_implementation=drop_implementation,
            return_global_import=return_global_import,
        )

    def get_class(
        self,
        class_name: str,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        return_global_import: bool = False,
    ) -> str:
        class_node = locate_class(ast_obj=self.node, class_name=class_name)
        return self._unparse(
            node=class_node,
            drop_directive=drop_directive,
            directive_prefix=directive_prefix,
            drop_implementation=drop_implementation,
            return_global_import=return_global_import,
        )

    def get_entity(
        self,
        entity_name: str,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        return_global_import: bool = False,
    ) -> str:
        entity_node: ast.AST = locate_entity(code=self.node, entity_name=entity_name)

        return self._unparse(
            node=entity_node,
            drop_directive=drop_directive,
            directive_prefix=directive_prefix,
            drop_implementation=drop_implementation,
            return_global_import=return_global_import,
        )

    def get_entity_map(
        self, path: List[str] = None, refresh: bool = False
    ) -> EntityMap:
        if (self._entity_map is None) or refresh:
            entity_map = {}

            # Extract relative import
            import_list = extract_import_list(
                self.node, 
                drop_directive=True, 
                relative_only=False
            )
            for import_node in import_list:
                if isinstance(import_node, ast.ImportFrom):
                    module_name = import_node.module
                    imported_items = [alias.name for alias in import_node.names]
                    module_path = resolve_module_filepath(
                        module_name,
                        relative_level=max([1, import_node.level]),
                        absolute_base_path=self.filename,
                    )
                    # Resolve filepath from module_name
                    for i in imported_items:
                        entity_map[i] = module_path
            self._entity_map = entity_map
            
            

            return entity_map

        else:
            return self._entity_map

    def unparse(
        self,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        exception: Optional[List[str]] = None,
        return_global_import: bool = False,
    ) -> str:
        return self._unparse(
            node=self.node,
            directive_prefix=directive_prefix,
            drop_directive=drop_directive,
            drop_implementation=drop_implementation,
            exception=exception,
            return_global_import=return_global_import,
        )

    def _unparse(
        self,
        node: ast.AST,
        drop_directive: bool = False,
        directive_prefix: str = "",
        drop_implementation: bool = False,
        exception: Optional[List[str]] = None,
        return_global_import: bool = False,
    ) -> str:
        # We need to create a copy to avoid mutation
        node = copy.deepcopy(node)
        if drop_directive:
            if directive_prefix == "":
                raise ValueError(
                    "Must specify a directive_prefix when drop_directive set to True"
                )
            transformer = DropDirective(prefix=directive_prefix)
            node = transformer.visit(node)

        if drop_implementation:
            transformer = DropImplementation(exception=exception if exception else [])
            node = transformer.visit(node)

        ast.fix_missing_locations(node)

        if return_global_import:
            return (
                self.get_import_statement(
                    drop_directive=drop_directive, directive_prefix=directive_prefix
                )
                + "\n\n"
                + ast_comments.unparse(node)
            )

        else:
            return ast_comments.unparse(node)

    def save(self, filename: Optional[str] = None) -> None:
        filename = filename if filename else self.filename
        write_file(filename, ast_comments.unparse(self.node))

    def serialize(self, output_file: str) -> None:
        with open(output_file, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_file(cls, filename: str) -> ASTSourceCode:
        content = read_file(filename)
        code = ast_comments.parse(content)
        return cls(node=code, filename=filename)

    @classmethod
    def from_pickle(cls, pickle_file: str) -> ASTSourceCode:
        with open(pickle_file, "rb") as f:
            return pickle.load(f)



import pkgutil

from langchain.prompts import PromptTemplate


def load_template(filename: str) -> PromptTemplate:
    coder_prompt_str = pkgutil.get_data(__name__, f"prompts/{filename}").decode("utf-8")
    coder_prompt = PromptTemplate.from_template(
        template=coder_prompt_str, template_format="jinja2"
    )
    return coder_prompt


CODER_PROMPT = load_template("coder.md")
TESTER_FUNCTION = load_template("write_test.md")



from contextlib import contextmanager
from pathlib import Path

from wander import Renderer
from wander.plugins import ShellPlugin


def wander_render(text: str) -> str:
    renderer = Renderer(plugins=[ShellPlugin()])
    return renderer.render(text)


@contextmanager
def temporary_workspace():
    """Context manager that creates a file when entering and deletes it when exiting."""
    path = get_workspace_file()
    path.touch()  # Create the file

    try:
        yield path  # Yield the path to the caller

    finally:
        path.unlink()  # Delete the file


def get_workspace_file() -> Path:
    """Returns a Path object representing an available workspace file path.

    The function checks for the existence of workspace files in the ".wildered" directory
    and returns a new file path that is not already in use. The file paths follow the pattern
    "workspace_X.md", where X is a unique number starting from 1.

    Returns:
        A Path object representing an available workspace file path.
    """
    base_path = Path(".wildered/")
    base_path.mkdir(exist_ok=True)
    file_prefix = "workspace"
    file_number = 1
    while (base_path / f"{file_prefix}_{file_number}.md").exists():
        file_number += 1
    return base_path / f"{file_prefix}_{file_number}.md"



from typing import Annotated, Optional

import typer

from ..commands import scan as _scan

app = typer.Typer()

SCAN_HELP = "Scan for autocomplete directives in the specified filename and output the result in a markdown file."
SCAN_CLIPBOARD_HELP = "Whether to copy the formatted prompt to your clipboard"
SCAN_REMOVE_DIRECTIVE_HELP = """\
Whether to remove the directive after the program exits. \
WARNING: This will mess up your code formatting as AST is used. \
In the future, we will allow manipulation of source code without messing up the formatting.\
"""
SCAN_AUTO_INTEGRATE = """\
Whether to integrate the LLM response automatically into your script.
WARNING: This will mess up your code formatting as AST is used. \
In the future, we will allow manipulation of source code without messing up the formatting.\
"""

@app.command(help=SCAN_HELP)
def scan(
    filename: str,
    clipboard: Annotated[bool, typer.Option(help=SCAN_CLIPBOARD_HELP, show_default="True")] = True,
    remove_directive: Annotated[bool, typer.Option(help=SCAN_REMOVE_DIRECTIVE_HELP, show_default="False")] = False,
    auto_integrate: Annotated[bool, typer.Option(help=SCAN_AUTO_INTEGRATE, show_default="False")] = False
):
    _scan(
        filename=filename,
        clipboard=clipboard,
        remove_directive=remove_directive,
        auto_integrate=auto_integrate
    )


def main():
    app()



import ast
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

from typing_extensions import assert_never


def get_call_name(call_node: ast.Call) -> str:
    if hasattr(call_node.func, "id"):
        return ("", call_node.func.id)

    else:
        return (call_node.func.value.id, call_node.func.attr)


# TODO: Split the function
def locate_import(
    imports: List[Union[ast.Import, ast.ImportFrom]], keyword: str
) -> Optional[Union[ast.Import, ast.ImportFrom]]:
    """Locates the import statement with the given `keyword`.
    Args:
        imports: A list of `ast.Import` or `ast.ImportFrom` nodes representing import statements.
        keyword: The name or module being searched.
    Returns:
        The matching `ast.Import` or `ast.ImportFrom` node if found, or `None` if not found.
    Raises:
        None.
    """
    for imp in imports:
        if isinstance(imp, ast.Import):
            for name in imp.names:
                if name.name == keyword:
                    return imp
                
        elif isinstance(imp, ast.ImportFrom):
            continue
        
        else:
            assert_never(imp)

    return None

def locate_import_from(imports: List[Union[ast.Import, ast.ImportFrom]], keyword: str, level: int
) -> Optional[ast.ImportFrom]:
    for imp in imports:
        if isinstance(imp, ast.Import):
            continue
        
        elif isinstance(imp, ast.ImportFrom):
            if imp.module == keyword and imp.level == level:
                return imp
        
        else:
            assert_never(imp)
            
    return None

def integrate_alias(old_aliases: List[ast.alias], new_aliases: List[ast.alias]) -> None:
    old_alias_names = [i.name for i in old_aliases]
    def is_new_alias(alias: ast.alias) -> bool:
        return not(alias.name in old_alias_names)

    new_alias_list = filter(is_new_alias, new_aliases)
    old_aliases.extend(new_alias_list)

def diff_import_list(
    old_imports: List[Union[ast.Import, ast.ImportFrom]],
    new_imports: List[Union[ast.Import, ast.ImportFrom]],
) -> List[Union[ast.Import, ast.ImportFrom]]:
    """Detects new or modified import statements in `new_imports` compared to `old_imports`.
    Args:
        old_imports: A list of `ast.Import` or `ast.ImportFrom` nodes representing the old import statements.
        new_imports: A list of `ast.Import` or `ast.ImportFrom` nodes representing the new import statements.
    Returns:
        A tuple containing two lists:
        - `new_import_ast`: A list of new import statements found in `new_imports`.
        - `modified_import_from_ast`: A list of modified `ast.ImportFrom` statements found in `new_imports`.
    Raises:
        None.
    """
    # Will modify the import from inplace if available.
    new_import_ast = []

    for i in new_imports:
        if isinstance(i, ast.Import):
            match = locate_import(old_imports, i.names[0].name)
            if match is None:  # Meaning it is a new import
                new_import_ast.append(i)
        elif isinstance(i, ast.ImportFrom):
            match = locate_import_from(old_imports, i.module, level=i.level)
            if match is None:  # New import
                new_import_ast.append(i)
            else:
                # TODO: Not elegant
                integrate_alias(old_aliases=match.names, new_aliases=i.names)
    return new_import_ast


class DropDirective(ast.NodeTransformer):
    def __init__(self, prefix: str):
        self.prefix = prefix

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        node.decorator_list = filter(self.filter_directive, node.decorator_list)
        self.generic_visit(node)
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        node.decorator_list = filter(self.filter_directive, node.decorator_list)
        self.generic_visit(node)
        return node

    def visit_Expr(self, node: ast.Expr) -> Optional[ast.Expr]:
        try:
            prefix, node_name = get_call_name(node.value)
            if (prefix == self.prefix) and (node_name == "run"):
                return None
            
            else:
                return node

        except AttributeError:
            self.generic_visit(node)
            return node

    def visit_Import(self, node: ast.Import) -> Optional[ast.Import]:
        if node.names[0].name.startswith(self.prefix):
            return None

        else:
            return node

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Optional[ast.ImportFrom]:
        if node.module.startswith(self.prefix):
            return None

        else:
            return node

    def filter_directive(self, decorator: Union[ast.Call, ast.Name]) -> bool:
        if isinstance(decorator, ast.Call):
            prefix, node_name = get_call_name(decorator)

            if prefix == self.prefix:
                return False

            return True

        else:
            return True


class ReplaceNode(ast.NodeTransformer):
    def __init__(
        self,
        class_to_replace: Dict[str, ast.ClassDef] = None,
        function_to_replace: Dict[str, ast.FunctionDef] = None,
        method_to_replace: Dict[str, Dict[str, ast.FunctionDef]] = None,
    ):
        # TODO: Nested method not supported
        self.class_to_replace = class_to_replace if class_to_replace else {}
        self.function_to_replace = function_to_replace if function_to_replace else {}
        self.method_to_replace = method_to_replace if method_to_replace else {}
        self.cur_class = ""

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        if node.name in self.function_to_replace:
            self.generic_visit(node)
            return self.function_to_replace[node.name]

        elif (self.cur_class in self.method_to_replace) and (
            node.name in self.method_to_replace[self.cur_class]
        ):
            self.generic_visit(node)
            return self.method_to_replace[self.cur_class][node.name]

        else:
            self.generic_visit(node)
            return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        if node.name in self.class_to_replace:
            self.generic_visit(node)
            return self.class_to_replace[node.name]

        else:
            # TODO: This will not work for nested
            self.cur_class = node.name
            self.generic_visit(node)
            self.cur_class = None
            return node


class DropImplementation(ast.NodeTransformer):
    def __init__(self, exception: List[str]) -> None:
        self.exception = exception

    def visit_FunctionDef(self, node: ast.FunctionDef):
        if node.name in self.exception:
            self.generic_visit(node)
            return node

        else:
            docstring = ast.get_docstring(node)
            if docstring:
                node.body = [ast.Expr(value=ast.Constant(value=docstring))]

            else:
                node.body = []

            return node


def locate_function(ast_obj: ast.AST, func_name: str) -> ast.FunctionDef:
    for node in ast.walk(ast_obj):
        if isinstance(node, ast.FunctionDef):
            if node.name == func_name:
                return node

    raise ValueError(f"Function with function name '{func_name}' does not exist")


def locate_class(ast_obj: ast.AST, class_name: str) -> ast.ClassDef:
    for node in ast.walk(ast_obj):
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            return node

    raise ValueError(f"Class with class name '{class_name}' does not exist")


def locate_method(ast_obj: ast.AST, func_name: str, class_name: str) -> ast.FunctionDef:
    class_node = locate_class(ast_obj, class_name=class_name)
    method_node = locate_function(class_node, func_name=func_name)
    return method_node


def is_directive_import(
    import_ast: Union[ast.Import, ast.ImportFrom], directive_prefix: str
) -> bool:
    if isinstance(import_ast, ast.Import):
        if import_ast.names[0].name == directive_prefix:
            return True

        else:
            return False

    elif isinstance(import_ast, ast.ImportFrom):
        if import_ast.module == directive_prefix:
            return True

        else:
            return False

    else:
        raise ValueError(f"Expected import ast, got type '{import_ast}' instead")


def extract_import_list(
    code: ast.AST,
    drop_directive: bool = False,
    directive_prefix: str = "",
    relative_only: bool = False,
) -> List[Union[ast.Import, ast.ImportFrom]]:
    """
    Extract all leading import
    """
    imports = []
    for node in ast.iter_child_nodes(code):
        if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
            if relative_only and not (
                isinstance(node, ast.ImportFrom) and node.level > 0
            ):
                continue

            elif drop_directive and is_directive_import(
                node, directive_prefix=directive_prefix
            ):
                continue

            else:
                imports.append(node)

        # Stop at the first non-import statement
        else:
            break

    return imports


class NodeFinder(ast.NodeVisitor):
    def __init__(
        self,
        entity_list: Optional[List[str]] = None,
    ) -> None:
        """
        Initialize the NodeFinder class.
        Args:
            entity_list: Optional[List[str]] = None
        """
        self.entity_list = entity_list if entity_list else []
        self.found_node = defaultdict(lambda: None)

    def visit_FunctionDef(self, node):
        if node.name in self.entity_list:
            self.found_node[node.name] = node

        self.generic_visit(node)
        
        return node

    def visit_ClassDef(self, node):
        if node.name in self.entity_list:
            self.found_node[node.name] = node

        self.generic_visit(node)
        
        return node


def locate_entity(code: ast.AST, entity_name: str):
    finder = NodeFinder(entity_name)
    finder.visit(code)
    return finder.found_node[entity_name]



from pathlib import Path


def resolve_module_filepath(
    module_name: str, relative_level: int, absolute_base_path: Path
) -> str:
    relative_path = ""
    for i in range(relative_level):
        relative_path += "../"

    relative_path += module_name
    return apply_relative_path(
        relative_path=Path(relative_path), absolute_base_path=absolute_base_path
    ).with_suffix(".py")


def apply_relative_path(relative_path: Path, absolute_base_path: Path) -> Path:
    """
    Applies a relative path to a base path, similar to the behavior of the 'cd' command in Linux.

    Args:
        base_path (Path): The base path to which the relative path will be applied.
        relative_path (Path): The relative path to be applied.

    Returns:
        Path: The resulting path after applying the relative path to the base path.

    """
    result_path = (
        absolute_base_path / relative_path
    )  # Apply the relative path to the base path
    return result_path.resolve()


def read_file(filepath: str) -> str:
    with open(filepath, "r") as f:
        text = f.read()
        return text


def write_file(filepath: str, content: str) -> None:
    filepath = Path(filepath)
    filepath.parent.mkdir(exist_ok=True, parents=True)
    with open(filepath, "w") as f:
        f.write(content)



from typing import Annotated, Callable, Sequence

from pydantic import BaseModel

from wildered.models import BaseEntity


class EntityGroup(BaseModel):
    group_name: str
    entity_list: Sequence[BaseEntity]


class EntityGrouper(BaseModel):
    group_func: Callable[[BaseEntity], Annotated[str, "Group name of the entity"]]

    class Config:
        arbitrary_types_allowed = True

    def group(self, entity_list: Sequence[BaseEntity]) -> Sequence[EntityGroup]:
        group_map = {}
        for entity in entity_list:
            group_name = self.group_func(entity)
            group_map.setdefault(group_name, [])
            group_map[group_name].append(entity)

        return [
            EntityGroup(group_name=name, entity_list=entity_list)
            for name, entity_list in group_map.items()
        ]



import ast
import inspect
from abc import ABC
from typing import (
    ClassVar,
    List,
    Literal,
    Optional,
    Set,
    TypeAlias,
)

import libcst as cst
from pydantic import BaseModel, Field, root_validator, validator
from pydantic.dataclasses import dataclass

DirectiveContext: TypeAlias = Literal["module", "class", "function"]


class Identifier(BaseModel):
    """A wrapper around variable-like names"""

    node: ast.Expr | ast.Name | cst.Expr | cst.Name
    name: str

    class Config:
        arbitrary_types_allowed = True


class BaseDirectiveConfig(object):
    name: str
    positional_order: Optional[List[str]] = None
    alias: Optional[List[str]] = None
    allow_multiple: bool = True
    allowed_context: Set[DirectiveContext] = Field(default_factory=set)  # TODO
    requires: List[str] = Field(default_factory=list)
    resists: List[str] = Field(default_factory=list)

    @validator("allowed_context", always=True)
    def set_default_context(cls, v):
        if not v:
            return {"module", "class", "function"}

        else:
            return v

    @root_validator()
    def set_default_alias(cls, values):
        if values["alias"] is None:
            values["alias"] = [values["name"]]

        return values

    @root_validator()
    def compatible_requires_conflict(cls, values):
        if values["requires"] and values["resists"]:
            for i in values["requires"]:
                if i in values["resists"]:
                    raise ValueError(f"{i} is in both requires and conflict list.")

        return values


def get_config_settings(obj: BaseDirectiveConfig):
    defined_settings = {
        property: value
        for property, value in vars(obj).items()
        if not property.endswith("__")
    }

    return defined_settings


def get_config_object(cls):
    return [
        cls_attribute
        for cls_attribute in cls.__dict__.values()
        if inspect.isclass(cls_attribute)
        and issubclass(cls_attribute, BaseDirectiveConfig)
    ]


DirectiveConfig = dataclass(BaseDirectiveConfig).__pydantic_model__


class Directive(BaseModel, ABC):
    """Base class for holding directives"""

    _name: str  # What is the actual directive name used
    class_config: ClassVar[Optional[DirectiveConfig]] = None

    @root_validator(pre=True)
    def handle_positional_args(cls, values: dict) -> dict:
        if values.get("args", None):
            positional_order = cls.config.positional_order
            if positional_order is None:
                raise ValueError(
                    f"Attempt to pass positional arguments for {cls.__name__} without specifying positional orders"
                )

            for i in range(len(values["args"])):
                try:
                    values[positional_order[i]] = values["args"][i]

                except IndexError:
                    raise ValueError("Positional arguments number exceed")

        return values

    @root_validator(pre=True)
    def validate_inner_directive(cls, values: dict) -> dict:
        config = get_config_object(cls)
        if len(config) == 0:
            raise ValueError("You must specify a config object. ")

        if len(config) > 1:
            raise ValueError("You should only specify one config object. ")

        return values

    @classmethod
    @property
    def config(cls) -> DirectiveConfig:
        if cls.class_config is None:
            config_obj = get_config_object(cls)
            settings = get_config_settings(config_obj[0])
            config = DirectiveConfig(**settings)
            cls.class_config = config
            return config

        else:
            return cls.class_config



from __future__ import annotations

import ast
import copy
from typing import (
    Annotated,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

import ast_comments
from pydantic import validator
from typing_extensions import assert_never

from wildered.ast.utils import (
    DropDirective,
    DropImplementation,
    get_call_name,
    locate_function,
)
from wildered.directive import Directive, DirectiveContext, Identifier
from wildered.models import BaseDirectiveParser, BaseEntity

from .source_code import ASTSourceCode
from .utils import (
    locate_class,
)


class ASTDirectiveParser(BaseDirectiveParser):
    def parse(
        self, source: ASTSourceCode, drop_directive: bool = True
    ) -> List[ASTEntity]:
        # Drop_directive will mutate the SourceCode
        detector = DetectDirective(
            parser=self,
            source=source,
            directive_list=self.directives,
            code=source.node,
            prefix=self.prefix_name,
        )

        detector.visit(source.node)

        if drop_directive:
            drop_transformer = DropDirective(prefix=self.prefix_name)
            drop_transformer.visit(source.node)

        self.check_valid_combination(entity_list=detector.detected)
        return detector.detected


class ASTEntity(BaseEntity):
    wrapping_node: Optional[ASTEntity]
    node: ast.AST
    name: str
    source: ASTSourceCode
    directives: Dict[str, List[Directive]]
    parser: ASTDirectiveParser
    _context: DirectiveContext

    class Config:
        arbitrary_types_allowed = True

    def save(self, *args, **kwargs) -> None:
        self.source.save(*args, **kwargs)

    def unparse(
        self,
        drop_directive: bool = False,
        drop_implementation: bool = False,
        return_global_import: bool = False,
        include_ancestor: bool = False,
    ) -> str:
        if include_ancestor:
            # Top ancestor
            node = copy.deepcopy(self.ancestor[-1].node)

        else:
            node = copy.deepcopy(self.node)

        if drop_directive:
            transformer = DropDirective(prefix=self.parser.prefix_name)
            transformer.visit(node)

        if drop_implementation:
            transformer = DropImplementation(exception=[])
            transformer.visit(node)

        ast.fix_missing_locations(node)

        if return_global_import:
            return (
                self.source.get_import_statement(
                    drop_directive=drop_directive,
                    directive_prefix=self.parser.prefix_name,
                )
                + "\n\n"
                + ast_comments.unparse(node)
            )

        else:
            return ast_comments.unparse(node)


class ASTClassEntity(ASTEntity):
    node: ast.ClassDef
    _context = "class"

    def update(
        self,
        new_code: ast.ClassDef | str,
        modify_signature: bool = True,
        name: Optional[str] = None,
    ) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)
            new_code = locate_class(new_code, class_name=name if name else self.name)

        self.node.body = new_code.body
        self.node.decorator_list = new_code.decorator_list
        
        if modify_signature:
            self.node.name = new_code.name
            self.node.bases = new_code.bases
            self.node.keywords = new_code.keywords


class ASTFunctionEntity(ASTEntity):
    node: ast.FunctionDef
    _context = "function"

    def update(
        self,
        new_code: ast.FunctionDef | str,
        modify_signature: bool = True,
        name: Optional[str] = None,
    ) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)
            new_code = locate_function(new_code, func_name=name if name else self.name)

        self.node.body = new_code.body
        self.node.decorator_list = new_code.decorator_list
        
        if modify_signature:
            self.node.name = new_code.name
            self.node.args = new_code.args
            self.node.returns = new_code.returns


class ASTModuleEntity(ASTEntity):
    _context = "module"

    @validator("wrapping_node")
    def check_wrapping_node(cls, v):
        if v is not None:
            raise ValueError(
                "ASTModuleEntity should not have wrapping_node, make sure the run directive is defined at top level."
            )

        return v

    def update(self, new_code: ast.Module | str) -> None:
        if isinstance(new_code, str):
            new_code = ast_comments.parse(new_code)

        self.source.node = new_code
        self.node = new_code


ASTDirectiveParser.update_forward_refs()


def parse_arguments(node: ast.Call) -> Tuple[List[Any], dict[str, Any]]:
    positional_arg = []
    keyword_arg = {}

    for arg in node.args:
        arg_value = extract_value(arg)
        positional_arg.append(arg_value)

    for keyword in node.keywords:
        arg_value = extract_value(keyword.value)
        keyword_arg[keyword.arg] = arg_value

    return positional_arg, keyword_arg


def extract_value(node: ast.AST):
    match node:
        case ast.Constant():
            return node.value

        case ast.Expr():
            return Identifier(node=node, name=node.value)

        case ast.Dict():
            return extract_dict(node=node)

        case ast.Name():
            return Identifier(node=node, name=node.id)

        case ast.Str():
            return node.s

        case ast.Num():
            return node.n

        case ast.List():
            return extract_list(node)

        case other:
            assert_never(other)


def extract_dict(node: ast.Dict):
    return {
        extract_value(key): extract_value(value)
        for key, value in zip(node.keys, node.values)
    }


def extract_list(node: ast.List):
    return [extract_value(elt) for elt in node.elts]


class DetectDirective(ast.NodeVisitor):
    # Go through the node, return a list of task
    def __init__(
        self,
        parser: ASTDirectiveParser,
        source: ASTSourceCode,
        directive_list: List[Type[Directive]],
        code: ast.AST,
        prefix: str,
    ):
        self.parser = parser
        self.code = code
        self.source = source
        self.prefix = prefix
        self.detected = []
        self.task_list = []
        self.wrapping_node = None

        self.directives = {}
        self.config_map = {}

        for directive_cls in directive_list:
            config = directive_cls.config
            alias = config.alias
            name = config.name
            self.directives[name] = {"alias": alias, "cls": directive_cls}
            self.config_map[name] = config

    def visit_Call(self, node: ast.Call) -> Any:
        try:
            prefix, node_name = get_call_name(node)
            if (prefix == self.prefix) and (node_name == "run"):
                directives = self.extract_directives(
                    decorators=node.args,
                )
                if len(directives) != 0:
                    node_task = ASTModuleEntity(
                        parser=self.parser,
                        source=self.source,
                        wrapping_node=self.wrapping_node,
                        node=self.code,
                        name=self.source.filename.name,
                        directives=directives,
                    )
                    self.detected.append(node_task)

        except AttributeError:
            pass

        finally:
            self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        directives = self.extract_directives(decorators=node.decorator_list)
        node_task = ASTFunctionEntity(
            parser=self.parser,
            source=self.source,
            wrapping_node=self.wrapping_node,
            node=node,
            name=node.name,
            directives=directives,
        )

        if len(directives) > 0:
            self.detected.append(node_task)

        previous = self.wrapping_node
        self.wrapping_node = node_task
        self.generic_visit(node=node)
        self.wrapping_node = previous

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        directives = self.extract_directives(
            decorators=node.decorator_list,
        )
        node_task = ASTClassEntity(
            parser=self.parser,
            source=self.source,
            wrapping_node=self.wrapping_node,
            node=node,
            name=node.name,
            directives=directives,
        )

        if len(directives) > 0:
            self.detected.append(node_task)

        previous = self.wrapping_node
        self.wrapping_node = node_task
        self.generic_visit(node=node)
        self.wrapping_node = previous

    def extract_directives(self, decorators: ast.AST) -> Dict[str, List[Directive]]:
        directive_map = {}
        for node in decorators:
            if isinstance(node, ast.Call):
                prefix_name, name = get_call_name(node)
                if prefix_name != self.prefix:
                    continue

                matched_directive = self.directive_lookup(name)
                if matched_directive is not None:
                    pos_args, keyword_args = parse_arguments(node)
                    directive = matched_directive[1](
                        args=pos_args, **keyword_args, _name=matched_directive[0]
                    )
                    directive_map.setdefault(matched_directive[2], [])
                    directive_map[matched_directive[2]].append(directive)

        return directive_map

    def directive_lookup(
        self, name: str
    ) -> (
        Tuple(
            Annotated[str, "The currently used alias"],
            Type[Directive],
            Annotated[str, "Name of the directive"],
        )
        | None
    ):
        for i, j in self.directives.items():
            aliases = j["alias"]
            for alias in aliases:
                if alias == name:
                    return (name, j["cls"], i)

        return None



import inspect
from pathlib import Path
from typing import Any, Dict, List, Literal

import pyperclip
import guidance

from wildered.context.utils import temporary_workspace
from wildered.utils import read_file, write_file
from wildered.logger import logger

from .tasks import (
    TaskGroup,
)

# Use OPENAI_API_KEY env
guidance.llm = guidance.llms.OpenAI("text-davinci-003")

def task_executor(
    task_list: List[TaskGroup],
    clipboard: bool = False,
    auto_integrate: bool = False,
) -> None:
    for group in task_list:
        final_prompt = format_task_prompt(group=group, clipboard=clipboard)
        
        if auto_integrate:
            final_prompt = augment_guidance_prompt(prompt=final_prompt)
            response = get_llm_response(final_prompt)
            logger.debug(f"LLM response: {response=}")
            group.integrate(response=response)

def format_task_prompt(group: TaskGroup, clipboard: bool) -> str:
    prompt = group.format_prompt()

    if clipboard:
        print("Copied prompt to clipboard")
        pyperclip.copy(prompt)

    with temporary_workspace() as f:
        write_file(f, prompt)
        print(f"Prompt wrote into {f}")
        print(f"Please replace it with the LLM responses.")
        _ = input("Press enter to exit. ")
        return read_file(f)


def augment_guidance_prompt(prompt: str) -> str:
    logger.debug(f"Before guidance: {prompt=}")
    additions = inspect.cleandoc("""
        Your answer should only consist of code. All explanation should be done with comments instead of raw text.
        ```python
        {{gen 'code'}}
        ```
    """)
    prompt = prompt + additions
    logger.debug(f"After guidance: {prompt=}")
    return prompt

def get_llm_response(prompt: str) -> str:
    guidance_program = guidance(prompt)
    raw_response = guidance_program()
    cleaned = raw_response['code'].strip("```")
    return cleaned


from typing import List, Optional

from pydantic import root_validator

from wildered.ast import ASTDirectiveParser
from wildered.directive import BaseDirectiveConfig, Directive, Identifier

from .utils import wander_render


class AutocompleteDirective(Directive):
    requirement: str = ""
    rewrite_docs: bool = False
    annotations: bool = True
    pseudo_code: bool = False
    group: Optional[str] = None
    modify: bool = False
    count: int = 1  # Useful for suggestions later

    class DirectiveConfig(BaseDirectiveConfig):
        name = "autocomplete"
        allow_multiple = False
        positional_order = ["requirement"]

    class Config:
        arbitrary_types_allowed = True


class HintDirective(Directive):
    entity_list: List[str | Identifier]
    infer: bool = False

    class Config:
        arbitrary_types_allowed = True

    class DirectiveConfig(BaseDirectiveConfig):
        name = "hint"
        positional_order = ["entity_list", "infer"]
        allow_multiple = True
        requires = {"autocomplete"}


butterfly_parser = ASTDirectiveParser(
    prefix_name="wildered", directives=[AutocompleteDirective, HintDirective]
)



from __future__ import annotations

import copy
from typing import Annotated, Any, Dict, List, Optional, Tuple, Type

import libcst as cst
from pydantic import validator
from typing_extensions import assert_never

from wildered.cst.source_code import CSTSourceCode
from wildered.cst.utils import (
    CSTDropDirective,
    get_call_name,
)
from wildered.directive import Directive, DirectiveContext, Identifier
from wildered.models import BaseDirectiveParser, BaseEntity


class CSTDirectiveParser(BaseDirectiveParser):
    def parse(self, source: CSTSourceCode, drop_directive: bool = True) -> List[CSTEntity]:
        detector = CSTDetectDirective(
                    parser=self,
                    source=source,
                    directive_list=self.directives,
                    code=source.node,
                    prefix=self.prefix_name,
                )
        
        source.node.visit(detector)

        if drop_directive:
            drop_transformer = CSTDropDirective(prefix=self.prefix_name)
            source.node.visit(drop_transformer)

        self.check_valid_combination(entity_list = detector.detected)
        return detector.detected
    
class CSTEntity(BaseEntity):
    wrapping_node: Optional[CSTEntity]
    node: Any
    name: str
    source: CSTSourceCode
    directives: Dict[str, List[Directive]]
    parser: CSTDirectiveParser
    _context: DirectiveContext
    
    def unparse(
        self,
        drop_directive: bool = False,
        drop_implementation: bool = False,
        return_global_import: bool = False,
        include_ancestor: bool = False,
    ) -> str:
        if include_ancestor:
            # Top ancestor
            node = copy.deepcopy(self.ancestor[-1].node)

        else:
            node = copy.deepcopy(self.node)

        return self.source._unparse(node=node, 
                                    drop_directive=drop_directive, 
                                    directive_prefix=self.parser.prefix_name, 
                                    return_global_import=return_global_import, 
                                    drop_implementation=drop_implementation)
    
class CSTFunctionEntity(CSTEntity):
    _context = "function"
    
    @validator("wrapping_node")
    def check_wrapping_node(cls, v):
        if v is not None:
            raise ValueError(
                "ASTModuleEntity should not have wrapping_node, make sure the run directive is defined at top level."
            )
            
    def update(self, new_code: str) -> None:
        pass
    
class CSTModuleEntity(CSTEntity):
    _context = "module"
    
    @validator("wrapping_node")
    def check_wrapping_node(cls, v):
        if v is not None:
            raise ValueError(
                "ASTModuleEntity should not have wrapping_node, make sure the run directive is defined at top level."
            )
        
        return v
    
    def update(self, new_code: str) -> None:
        pass
    
class CSTClassEntity(CSTEntity):
    _context = "class"
    
    
    def update(self, new_code: str) -> None:
        pass

class CSTDetectDirective(cst.CSTVisitor):
    # Go through the node, return a list of task
    def __init__(
        self,
        parser: CSTDirectiveParser,
        source: CSTSourceCode,
        directive_list: List[Type[Directive]],
        code: cst.Module,
        prefix: str,
    ):
        self.parser = parser
        self.code = code
        self.source = source
        self.prefix = prefix
        self.detected = []
        self.task_list = []
        self.wrapping_node = None

        self.directives = {}
        self.config_map = {}

        for directive_cls in directive_list:
            config = directive_cls.config
            alias = config.alias
            name = config.name
            self.directives[name] = {"alias": alias, "cls": directive_cls}
            self.config_map[name] = config

    def visit_Call(self, node: cst.Call) -> Any:
        try:
            prefix, node_name = get_call_name(node)
            if (prefix == self.prefix) and (node_name == "run"):
                directives = self.extract_directives(
                    decorators=node.args,
                )
                if len(directives) != 0:
                    node_task = CSTModuleEntity(
                        parser=self.parser,
                        source=self.source,
                        wrapping_node=self.wrapping_node,
                        node=self.code,
                        name=self.source.filename.name,
                        directives=directives,
                    )
                    self.detected.append(node_task)

        except AttributeError:
            pass

    def visit_FunctionDef(self, node: cst.FunctionDef) -> Any:
        directives = self.extract_directives(
            decorators=node.decorators
        )
        node_task = CSTFunctionEntity(
            parser=self.parser,
            source=self.source,
            wrapping_node=self.wrapping_node,
            node=node,
            name=node.name.value,
            directives=directives,
        )

        if len(directives) > 0:
            self.detected.append(node_task)

        previous = self.wrapping_node
        self.wrapping_node = node_task
        self.wrapping_node = previous

    def visit_ClassDef(self, node: cst.ClassDef) -> Any:
        directives = self.extract_directives(
            decorators=node.decorators,
        )
        node_task = CSTClassEntity(
            parser=self.parser,
            source=self.source,
            wrapping_node=self.wrapping_node,
            node=node,
            name=node.name.value,
            directives=directives,
        )

        if len(directives) > 0:
            self.detected.append(node_task)

        previous = self.wrapping_node
        self.wrapping_node = node_task
        self.wrapping_node = previous

    def extract_directives(
        self, decorators: List[cst.Decorator]
    ) -> Dict[str, List[Directive]]:
        directive_map = {}
        [decorator.decorator for decorator in decorators]
        for node in decorators:
            if isinstance(node, cst.Call):
                prefix_name, name = get_call_name(node)
                if prefix_name != self.prefix:
                    continue

                matched_directive = self.directive_lookup(name)
                if matched_directive is not None:
                    pos_args, keyword_args = parse_arguments(node)
                    directive = matched_directive[1](
                        args=pos_args, **keyword_args, _name=matched_directive[0]
                    )
                    directive_map.setdefault(matched_directive[2], [])
                    directive_map[matched_directive[2]].append(directive)

        return directive_map

    def directive_lookup(
        self, name: str
    ) -> (
        Tuple(
            Annotated[str, "The currently used alias"],
            Type[Directive],
            Annotated[str, "Name of the directive"],
        )
        | None
    ):
        for i, j in self.directives.items():
            aliases = j["alias"]
            for alias in aliases:
                if alias == name:
                    return (name, j["cls"], i)

        return None


def parse_arguments(node: cst.Call) -> Tuple[List[Any], dict[str, Any]]:
    positional_arg = []
    keyword_arg = {}

    for arg in node.args:
        arg = arg.value
        if hasattr(arg, "keyword"):
            arg_name = arg.keyword.value
            arg_value = extract_value(arg.value.value)
            keyword_arg[arg_name] = arg_value
        
        else:
            arg_value = extract_value(arg.value.value)
            positional_arg.append(arg_value)

    return positional_arg, keyword_arg


def extract_value(node: cst.CSTNode):
    match node:       
        case cst.Expr():
            return Identifier(node=node, name=node.value)
        
        case cst.Dict():
            return extract_dict(node=node)
        
        case cst.Name():
            return Identifier(node=node, name=node.value)
        
        case cst.SimpleString():
            return node.value
        
        case cst.Float():
            return float(node.value)
        
        case cst.Integer():
            return int(node.value)
        
        case cst.List():
            return extract_list(node)
        
        case other:
            assert_never(other)

def extract_dict(node: cst.Dict) -> dict:
    result = {}
    for i in node.elements:
        key = i.key.value
        value = extract_value(i.value)
        result[key] = value
    
    return result


def extract_list(node: cst.List) -> list:
    return [extract_value(elt.value) for elt in node.elements]


from __future__ import annotations

from abc import ABC, abstractclassmethod
from typing import Any, Dict, List, Optional, Sequence, Type

from pydantic import BaseModel, validator

from wildered.directive import Directive, DirectiveConfig, DirectiveContext


class BaseSourceCode(BaseModel, ABC):
    class Config:
        arbitrary_types_allowed = True

    @abstractclassmethod
    def save(self, filename: str):
        raise NotImplementedError

    @abstractclassmethod
    def from_file(cls, filename: str):
        pass


class BaseEntity(BaseModel, ABC):
    wrapping_node: Optional[BaseEntity]
    node: Any
    name: str
    source: BaseSourceCode
    directives: Dict[str, List[Directive]]
    parser: BaseDirectiveParser
    _context: DirectiveContext

    class Config:
        arbitrary_types_allowed = True

    def save(self, *args, **kwargs) -> None:
        self.source.save(*args, **kwargs)

    @abstractclassmethod
    def unparse(
        self,
        drop_directive: bool = False,
        drop_implementation: bool = False,
        return_global_import: bool = False,
        include_ancestor: bool = False,
    ) -> str:
        raise NotImplementedError

    @property
    def ancestor(self) -> List[BaseEntity]:
        ancestor_list = []
        cur_node = self
        while cur_node.wrapping_node is not None:
            ancestor_list.append(cur_node.wrapping_node)
            cur_node = cur_node.wrapping_node

        return ancestor_list


class BaseDirectiveParser(BaseModel, ABC):
    prefix_name: str
    directives: List[Type[Directive]]

    class Config:
        arbitrary_types_allowed = True

    @validator("directives")
    def check_uniqueness(cls, v: List[Type[Directive]]) -> List[Type[Directive]]:
        existing_name = []
        for directive_type in v:
            name = directive_type.config.name
            if name not in existing_name:
                existing_name.append(name)

            else:
                raise ValueError(
                    f"Name '{name}' has already been used. All directives must have unique name."
                )

        return v

    def check_valid_context(
        self, directives: Dict[str, List[Directive]], context: DirectiveContext
    ) -> None:
        for directive_list in directives.values():
            if len(directive_list) > 0:
                directive_config: DirectiveConfig = type(directive_list[0]).config
                allowed_context = directive_config.allowed_context
                if context not in allowed_context:
                    raise ValueError(
                        f"Directive {directive_config.name} cannot be placed in {context}. "
                    )

                if len(directive_list) > 1:
                    allow_multiple = directive_config.allow_multiple
                    if not allow_multiple:
                        raise ValueError(
                            f"Directive {directive_config.name} does not allow multiple instances. "
                        )

                required = directive_config.requires
                for name in required:
                    if name not in directives.keys():
                        raise ValueError(
                            f"Directive {directive_config.name} requires directives {name}"
                        )

                resisted = directive_config.resists
                for name in resisted:
                    if name in directives.keys():
                        raise ValueError(
                            f"Directive {directive_config.name} resists directives {name}"
                        )

    def check_valid_combination(self, entity_list: Sequence[BaseEntity]) -> None:
        for entity in entity_list:
            context = entity._context
            self.check_valid_context(directives=entity.directives, context=context)
